@@ -1,3 +1,4 @@
+import importlib.util
 import logging
 import pathlib
 import typing
@@ -9,12 +10,21 @@ from pydantic import AnyHttpUrl
 from pydantic import BaseModel
 from pydantic import BaseSettings
 from pydantic import Extra
+from pydantic import validator
 from sentry_sdk.integrations.logging import LoggingIntegration
 
 from . import __version__
 from .dict_merge import dict_merge
 from .logging import setup_logging
 
+if typing.TYPE_CHECKING:
+    from .proxy import QueryProxy
+
+    QueryProxyType = typing.TypeVar('QueryProxyType', bound=QueryProxy)
+else:
+    QueryProxyType = None
+
+
 __all__ = (
     'NetworkModel',
     'ServerModel',
@@ -34,27 +44,73 @@ class NetworkModel(BaseModel):
     server_port: int
     bind_ip: IPv4Address
     bind_port: int
+    ebpf_no_redirect: bool = False
 
     class Config:
         extra = Extra.forbid
 
 
+class EntrypointModel(BaseModel):
+    """Allow override proxy server entrypoint
+    """
+
+    path: str  #: path to callable which should return :class:`QueryProxy`-based object
+    obj: typing.Optional[typing.Callable[..., QueryProxyType]] = None
+
+    class Config:
+        extra = Extra.forbid
+
+    @validator('obj', always=True)
+    def _import_obj(cls, v, values):
+        if v is not None:
+            return v
+
+        path: str = values['path']
+        if ':' not in path:
+            raise ValueError(f'entry not found, expected path(s): example.py:MyObj, example:MyObj. Given: {path}')
+
+        file_path, _, attr_name = path.rpartition(':')
+
+        if not file_path.isidentifier():
+            file_path = pathlib.Path(file_path).resolve()
+            spec = importlib.util.spec_from_file_location('entrypoint', file_path)
+            module = importlib.util.module_from_spec(spec)
+            module.__loader__.exec_module(module)  # noqa
+        else:
+            module = importlib.import_module(file_path)
+
+        try:
+            return getattr(module, attr_name)
+        except AttributeError as exc:
+            raise AttributeError(exc.args[0].replace("'entrypoint'", f"'{file_path}'"))
+
+
 class ServerModel(BaseModel):
+    meta: dict
     network: NetworkModel
     a2s_info_cache_lifetime: int = 5
     a2s_rules_cache_lifetime: int = 5
     a2s_players_cache_lifetime: int = 1
     src_query_port_lifetime: int = 10
     no_a2s_rules: bool = False
+    entrypoint: typing.Optional[EntrypointModel] = None
+
+    @validator('entrypoint', pre=True)
+    def _entrypoint(cls, v, values):
+        if isinstance(v, str):
+            v = {'path': v}
+
+        v['path'] = v['path'].format_map(values['meta'])
+        return v
 
     class Config:
-        extra = Extra.forbid
+        extra = Extra.allow
 
 
 class EBPFModel(BaseModel):
     enabled: bool = False
-    executable: str = 'python2'
-    script_path: pathlib.Path = './src-ebpf/redirect.py'
+    executable: typing.Union[str, typing.List[str]] = 'python2'
+    script_path: typing.Optional[pathlib.Path] = None
 
     class Config:
         extra = Extra.forbid
@@ -64,9 +120,8 @@ class Settings(BaseSettings):
     sentry_dsn: typing.Optional[AnyHttpUrl] = None
     confdir_0: pathlib.Path = '/etc/sqproxy/conf.d/'
     confdir_1: pathlib.Path = './conf.d/'
-    error_log: pathlib.Path = '/var/log/sqproxy/error.log'
-    debug_log: pathlib.Path = '/var/log/sqproxy/debug.log'
-    debug_log_enabled: bool = False
+    error_log: pathlib.Path = '/dev/null'
+    loglevel: str = 'INFO'
     piddir: typing.Optional[pathlib.Path] = None
 
     class Config:
@@ -76,6 +131,13 @@ class Settings(BaseSettings):
     def get_merged_config_data(self):
         return load_configs(iter_config_files(self.confdir_0, self.confdir_1))
 
+    @validator('loglevel')
+    def _check_loglevel(cls, v):
+        from logging import _checkLevel  # noqa
+
+        _checkLevel(v)
+        return v
+
 
 def _apply_defaults(target, defaults):
     target.update(dict_merge(defaults, target))
@@ -137,6 +199,10 @@ def load_configs(paths: typing.Iterable[pathlib.Path]):
                 global_defaults.append(config_defaults)
 
             if config:
+                meta = {'CONFPATH': path, 'CONFDIR': path.parent}
+                for server in (config.get('servers') or {}).values():
+                    server['meta'] = meta
+
                 configs.append(config)
 
     whole_config = {}
@@ -163,11 +229,7 @@ def setup(settings_: Settings = None):
             dsn=settings.sentry_dsn, integrations=[sentry_logging], release=__version__,
         )
 
-    debug_filename = settings.debug_log_enabled and settings.debug_log.as_posix() or None
-
-    setup_logging(
-        debug_filename=debug_filename, error_filename=settings.error_log.as_posix(),
-    )
+    setup_logging(error_filename=settings.error_log.as_posix(),)
 
     if settings.sentry_dsn:
         logger.debug('Sentry enabled')
